@import 'bootstrap/dist/css/bootstrap.min.css';

.App {
  text-align: center;
  display: flex;
}

.App-logo {
  height: 5vmin;
  pointer-events: none;
}

@media (prefers-reduced-motion: no-preference) {
  .App-logo {
    animation: App-logo-spin infinite 20s linear;
  }
}

.App-header {
  background-color: #282c34;
  min-height: 100vh;
  display: flex;
  flex-direction: column;
  align-items: center;
  justify-content: center;
  font-size: calc(10px + 2vmin);
  color: white;
}

.App-link {
  color: #61dafb;
}

@keyframes App-logo-spin {
  from {
    transform: rotate(0deg);
  }
  to {
    transform: rotate(360deg);
  }
}


/* Table styles */
.table {
  width: 100%;
  border-collapse: collapse;
  margin: 25px 0;
  font-size: 0.9em;
  font-family: sans-serif;
  min-width: 400px;
  box-shadow: 0 0 20px rgba(0, 0, 0, 0.15);
}

.table thead th {
  background-color: #009879;
  color: #ffffff;
  text-align: left;
}


.table th:hover {
  text-decoration: underline;
}

.table th,
.table td {
  padding: 12px 15px;
}

.table tbody tr {
  border-bottom: 1px solid #dddddd;
}

.table tbody tr:nth-of-type(even) {
  background-color: #f3f3f3;
}

.table tbody tr:last-of-type {
  border-bottom: 2px solid #009879;
}

.table tbody tr.active-row {
  font-weight: bold;
  color: #009879;
}

/* Style for Search Container */
.search-container {
  margin-bottom: 10px;
  text-align: right; /* Aligns the search input to the right */
}

.search-input {
  padding: 8px;
  font-size: 12px;
  border: 2px solid #ccc;
  border-radius: 4px;
  margin-right: 10px;
}

/* Stylish Buttons */
.button {
  background-color: #4CAF50; /* Green */
  border: none;
  color: white;
  padding: 10px 20px;
  text-align: center;
  text-decoration: none;
  display: inline-block;
  font-size: 16px;
  margin: 4px 2px;
  transition-duration: 0.4s;
  cursor: pointer;
  border-radius: 5px;
}

.button:hover {
  background-color: white;
  color: black;
  border: 2px solid #4CAF50;
}

/* You can add more specific styles for each button type if needed */
.button-start {
  background-color: #008CBA; /* Blue */
}

.button-start:hover {
  border: 2px solid #008CBA;
}

.button-stop {
  background-color: #f44336; /* Red */
}

.button-stop:hover {
  border: 2px solid #f44336;
}

.button-adjust {
  background-color: #e7e7e7; /* Light Grey */
  color: black;
}

.button-adjust:hover {
  border: 2px solid #e7e7e7;
}



.workgraph-item {
  display: flex;
  flex-grow: 1;
  margin-left: 250px; /* Same as sidebar width */
}
