WIDGET_INSTALLATION_MESSAGE = (
    "Widget dependency not found. To visualize the workgraph, please install the widget dependency. "
    "Use 'pip install aiida-workgraph[widget]' if installing from PyPI. "
    "For local source installations, use 'pip install .[widget]' and then build the JavaScript library. "
    "Refer to the documentation for more details."
)
