/* WorkGraph.css */

/* Pagination styles */
.pagination {
    display: flex;
    list-style-type: none;
    justify-content: center;
    padding: 0;
  }

  .pagination li {
    margin: 0 5px;
  }

  .pagination li a {
    border: 1px solid #007bff;
    padding: 5px 10px;
    text-decoration: none;
    color: #007bff;
    border-radius: 5px;
  }

  .pagination li a:hover {
    background-color: #007bff;
    color: white;
  }

  .pagination .active a {
    background-color: #007bff;
    color: white;
  }

.action-button {
    background-color: #007BFF; /* Set the background color */
    color: #fff; /* Set the text color */
    border: none;
    border-radius: 1px;
    cursor: pointer;
    margin-right: 16px; /* Add some spacing between buttons */
    padding: 4px 4px;
    font-size: 10px; /* Adjust the font size */
}

.action-button:hover {
    background-color: #0056b3; /* Change the background color on hover */
}

.table {
  width: 100%; /* Set the table width to 100% */
}
