.fill-area {
  display: table;
  z-index: -1;
  position: absolute;
  top: -320000px;
  left: -320000px;
  width: 640000px;
  height: 640000px;
}

.background {
  background-color: #ffffff;
  opacity: 1;
  background-image: linear-gradient(#f1f1f1 3.2px, transparent 3.2px),
    linear-gradient(90deg, #f1f1f1 3.2px, transparent 3.2px),
    linear-gradient(#f1f1f1 1.6px, transparent 1.6px),
    linear-gradient(90deg, #f1f1f1 1.6px, #ffffff 1.6px);
  background-size: 80px 80px, 80px 80px, 16px 16px, 16px 16px;
  background-position: -3.2px -3.2px, -3.2px -3.2px, -1.6px -1.6px,
    -1.6px -1.6px;
}
